

# Suppress all logging from canopen. TODO: Filter the messages to keep the ones we want to worry about.
# Note: there are no CRITICAL logging messages in canopen 2.4.1, so this should disable all sdo.client/server messages.
import logging
logging.getLogger("canopen.sdo.client").setLevel(logging.CRITICAL)
logging.getLogger("canopen.sdo.server").setLevel(logging.CRITICAL)


import os
import subprocess
import time
import hashlib
from filelock import FileLock, Timeout as FileLockTimeout
from feathersdk.utils.common import timediff, currtime
from feathersdk.utils.logger import debug, info
from feathersdk.utils.feathertypes import *


_LIB_DIR = os.path.dirname(os.path.abspath(__file__))
_LIB_SRC = os.path.join(_LIB_DIR, "socketcan_tcp.c")
_LIB_SO = os.path.join(_LIB_DIR, "libsocketcantcp.so")
_LIB_HASH_FILE = os.path.join(_LIB_DIR, "sct_source_hash.h")
_LIB_LOCK = os.path.join(_LIB_DIR, ".libsocketcantcp.so.lock")
_LIB_COMPILE_TIMEOUT_SECONDS = float(os.getenv("SOCKETCAN_COMPILE_TIMEOUT", "3.0"))

_START_SO_HASH_STR = b"__FEATHER_SOURCE_HASH__"
_END_SO_HASH_STR = b"__END_FEATHER_SOURCE_HASH__"


class CompilationFailedError(Exception):
    pass


def _get_files_to_hash() -> List[str]:
    path = os.path.dirname(__file__)
    valid_file = lambda f: f.endswith((".h", ".c")) and os.path.join(path, f) != _LIB_HASH_FILE
    return [os.path.join(path, f) for f in os.listdir(path) if valid_file(f)]


def _hash_file(file: str) -> str:
    """Eventually, might add more normalization to the file like removing comments"""
    with open(file, 'r') as f:
        data = f.read()
    
    hasher = hashlib.md5()
    hasher.update(data.encode("utf-8"))
    return hasher.hexdigest()


def _compile_lib(hash_str: str) -> None:
    """Compiles the library. Assumes we have file lock and are free to modify files"""
    info(f"Recompiling the socketcan_tcp library with new hash value: {hash_str}")
    
    # Place the hash val in the source code
    with open(_LIB_HASH_FILE, 'rb') as f:
        data = f.read()

    start, rest = data.split(_START_SO_HASH_STR)
    _, end = rest.split(_END_SO_HASH_STR)
    data = start + _START_SO_HASH_STR + hash_str.encode("ascii") + _END_SO_HASH_STR + end

    with open(_LIB_HASH_FILE, 'wb') as f:
        f.write(data)

    # Allow environment variable to override compile flags (useful for CI)
    compile_flags = os.getenv("SOCKETCAN_COMPILE_FLAGS", "-O3 -march=native -flto")
    cmd = ["gcc", "-fvisibility=hidden", "-shared"] + compile_flags.split() + ["-fPIC", "-o", _LIB_SO, _LIB_SRC]

    debug(f"Compiling library {_LIB_SO} with command: ", " ".join(cmd))
    try:
        subprocess.run(cmd, check=True, timeout=_LIB_COMPILE_TIMEOUT_SECONDS)
        if not os.path.exists(_LIB_SO):
            raise CompilationFailedError(f"Library {_LIB_SO} didn't exist after compilation")
    except Exception as e:
        raise e
    info(f"Library {_LIB_SO} compiled successfully")


def _needs_recompile(source_hash: str) -> bool:
    """Returns true if the file doesn't exist, or if source has changed since last recompile"""
    if not os.path.exists(_LIB_SO):
        debug(f"Library {_LIB_SO} doesn't exist, recompiling")
        return True

    if (so_hash_str := _get_binary_stored_hash()) is None:
        debug(f"Library {_LIB_SO} doesn't have hash string")
        return True

    debug(f"Library {_LIB_SO} hash value: {so_hash_str}, new hash value: {source_hash}")
    return source_hash != so_hash_str


def _get_binary_stored_hash() -> str:
    """Returns the hash string stored in the binary, or None if we couldn't find it"""
    with open(_LIB_SO, 'rb') as f:
        data = f.read()
    if _START_SO_HASH_STR not in data or _END_SO_HASH_STR not in data:
        return None
    
    return data.split(_START_SO_HASH_STR)[1].split(_END_SO_HASH_STR)[0].decode("ascii")


def _get_source_hash() -> str:
    """Returns the hash of the current source code"""
    return str(sum(int(_hash_file(f), 16) for f in _get_files_to_hash()))
        

def recompile_if_needed() -> str:
    """Recompiles library if it doesn't exist or there were changes to the file, then returns the filepath to .so:"""
    try:
        with FileLock(_LIB_LOCK, timeout=_LIB_COMPILE_TIMEOUT_SECONDS, mode=0o666):
            source_hash = _get_source_hash()
            if _needs_recompile(source_hash):
                _compile_lib(source_hash)
        
    except FileLockTimeout:
        # If we couldn't acquire the lock, another process is compiling. Wait for the library to appear and be complete
        start_time = currtime()
        while timediff(start_time) < _LIB_COMPILE_TIMEOUT_SECONDS:
            if os.path.exists(_LIB_SO):
                # Check if file size is stable (not being written)
                size1 = os.path.getsize(_LIB_SO)
                time.sleep(0.01)
                size2 = os.path.getsize(_LIB_SO)
                if size1 == size2 and size1 > 0:
                    break
            time.sleep(0.1)
        else:
            raise CompilationFailedError(f"Library was not compiled within {_LIB_COMPILE_TIMEOUT_SECONDS} seconds")
    
    return _LIB_SO



import os
import psutil
import signal
from filelock import FileLock, Timeout as _FLTimeoutError
import time
from feathersdk.utils.common import timediff, currtime
from feathersdk.utils.logger import info, debug, warning


LOCK_FILE_PATH = "/feathersys/user_config/.robot_instance.lock"
PID_FILE_PATH = "/feathersys/user_config/.robot_instance.pid"
LOCK_TIMEOUT_SECONDS = 1.0
SIGTERM_TIMEOUT_SECONDS = 2.0
_FILE_LOCK_MODE = 0o666  # Need to explicitly set global r/w in case we're running as root, and non-root is checking

__LOCK_ACQUIRED = False

# Makes sure that if you fork this process, the lock is reset
def __onfork_reset_lock_acquired() -> None:
    global __LOCK_ACQUIRED
    __LOCK_ACQUIRED = False
os.register_at_fork(after_in_child=__onfork_reset_lock_acquired)


class InstanceLockError(Exception):
    """Base class for all instance lock errors."""
    pass

class FileLockTimeoutError(InstanceLockError):
    """Raised when the file lock times out."""
    def __init__(self, lock_file_path: str):
        self.lock_file_path = lock_file_path
        super().__init__(f"Failed to acquire file lock: {lock_file_path}. Check no other processes are running " \
                         f"with Robot() instances, and remove the file with `rm {lock_file_path}`.")

class AlreadyHaveInstanceLockError(InstanceLockError):
    """Raised when the robot instance lock is already held by the current process."""
    def __init__(self):
        super().__init__("We already have the instance lock")

class DidNotHaveInstanceLockError(InstanceLockError):
    """Raised when the robot instance lock is not held by the current process."""
    def __init__(self):
        super().__init__("We do not have the instance lock")

class InvalidPIDFileError(InstanceLockError):
    """Raised when the PID in the PID file is invalid."""
    def __init__(self, pid_file_path: str):
        self.pid_file_path = pid_file_path
        super().__init__(f"Failed to parse PID from PID file: {pid_file_path}. Check no other processes are running " \
                         f"with Robot() instances, and remove the file with `rm {pid_file_path}`.")

class FailedToAcquireInstanceLockError(InstanceLockError):
    """Raised when the instance lock cannot be obtained because another process has it."""
    def __init__(self, instance_pid: int, can_kill: bool, override_lock: bool):
        self.instance_pid = instance_pid
        self.can_kill = can_kill
        self.override_lock = override_lock
        cannot_kill_msg = "" if can_kill else "\nNOTE: This process wouldn't have had permissions to kill the " \
            "previous instance. You'd have to either kill it manually or run this process as sudo."
        super().__init__(f"Another instance of the robot is already running. Use override_lock=True to override the " \
                         f"lock and kill the previous instance, or run `sudo kill -9 {instance_pid}` to kill the " \
                         f"previous instance. Override lock was \"{override_lock}\".{cannot_kill_msg}")

class KillPermissionsError(InstanceLockError):
    """Raised when the current process doesn't have permissions to kill the previous instance."""
    def __init__(self, instance_pid: int):
        self.instance_pid = instance_pid
        super().__init__(f"This process doesn't have permissions to kill the previous instance {instance_pid}. You'd " \
                         "have to either kill it manually or run this process as sudo.")


def get_robot_instance_lock(
    override_lock: bool = False, 
    strict: bool = True, 
    timeout: float = LOCK_TIMEOUT_SECONDS
) -> None:
    """Get the lock for the robot instance.

    Ensures that only one instance of the robot is running at a time. Does this by creating a file lock in the
    /feathersys/user_config directory. Lock file is named ".robot_instance.lock", and obtaining the lock allows a
    process to interact with the PID_FILE_PATH file which contains the PID of the running instance.

    If the pid file exists but the process is not running, the pid file will be removed and we will write our own PID
    to it. If the process is still running, we will raise an error (unless override_lock is True, in which case we will 
    kill the previous instance then write our own PID to the pid file).

    NOTE: You will require correct permissions to kill the previous instance. Either sudo, or the current user must
    have permissions to kill the previous instance.

    Args:
        override_lock: If True, will override the lock, killing the previous instance and allowing a new one to run. If
            False, will raise an error if the previous instance is still running.
        strict: If True, will raise an error if we attempt to acquire the lock while we already have it. Otherwise,
            doing so would succeed quietly.
        timeout: The timeout in seconds to wait for the file lock to be acquired.
    """
    os.makedirs(os.path.dirname(LOCK_FILE_PATH), exist_ok=True)
    
    global __LOCK_ACQUIRED
    if __LOCK_ACQUIRED:
        if strict:
            raise AlreadyHaveInstanceLockError()
        else:
            debug("Attempted to acquire instance lock that we already have. Returning early")
            return

    # In case we ever decide to change what we write to the PID file, this is a helper function to do so
    def _write_pid_file() -> None:
        global __LOCK_ACQUIRED
        with open(PID_FILE_PATH, 'w') as f:
            f.write(str(os.getpid()))
        __LOCK_ACQUIRED = True
        info("Successfully acquired robot instance lock")

    # Attempt to gain the lock by creating the lock file with current PID
    try:
        with FileLock(LOCK_FILE_PATH, timeout=timeout, mode=_FILE_LOCK_MODE):
            if not os.path.exists(PID_FILE_PATH):
                debug(f"No PID file {PID_FILE_PATH} found, writing our own PID")
                _write_pid_file()
                return
            
            # Otherwise, read the PID to see if that process is still running
            pid = _parse_pid_from_pid_file()
            
            # If the process is not running, remove the PID file and replace it with our own. Doesn't need special perms
            if not _is_process_running(pid):
                os.remove(PID_FILE_PATH)
                debug(f"Removed PID file {PID_FILE_PATH} because previous instance {pid} is not running")
                _write_pid_file()
                return
            
            # If the process is us, raise an error
            if pid == os.getpid():
                if strict:
                    raise AlreadyHaveInstanceLockError()
                __LOCK_ACQUIRED = True
                warning("somehow __LOCK_ACQUIRED was false but we own the lock!")
                return
            
            # Check if we have permissions to kill the previous instance
            try:
                os.kill(pid, 0)
                can_kill = True
            except (OSError, PermissionError):
                can_kill = False
            
            # The process is still running. If override_lock is False, raise an error
            if not override_lock:
                raise FailedToAcquireInstanceLockError(pid, can_kill, override_lock)
            
            # If we can't kill the previous instance, raise an error
            if not can_kill:
                raise KillPermissionsError(pid)
            
            # We can kill the previous instance. Attempt to send a SIGTERM
            os.kill(pid, signal.SIGTERM)
            debug(f"Sent SIGTERM to previous instance {pid}, waiting up to {SIGTERM_TIMEOUT_SECONDS} seconds to exit")

            init_time = currtime()
            while timediff(init_time) < SIGTERM_TIMEOUT_SECONDS:
                if not _is_process_running(pid):
                    debug(f"Previous instance {pid} has exited")
                    break
                time.sleep(0.01)
            else:
                # Sigterm timed out, send SIGKILL
                debug(f"Previous instance {pid} did not exit within {SIGTERM_TIMEOUT_SECONDS} seconds, sending SIGKILL")
                os.kill(pid, signal.SIGKILL)
                # Wait a bit for the process to become a zombie, then check
                time.sleep(0.1)
            
            # Make sure previous instance is actually dead (not just a zombie)
            if _is_process_running(pid):
                # Process is still running (not a zombie), which shouldn't happen after SIGKILL
                raise RuntimeError(f"Previous instance {pid} is somehow still alive after sigkill")
            else:
                debug(f"Previous instance {pid} has been terminated")
            
            # Write our own PID to the pid file
            _write_pid_file()
    
    except _FLTimeoutError:
        raise FileLockTimeoutError(LOCK_FILE_PATH)


def release_robot_instance_lock(allow_no_lock: bool = False, timeout: float = LOCK_TIMEOUT_SECONDS) -> None:
    """Release the robot instance lock

    Args:
        allow_no_lock: If True, will not raise an error if we don't have the lock.
        timeout: The timeout in seconds to wait for the file lock to be acquired.
    """
    global __LOCK_ACQUIRED
    if not __LOCK_ACQUIRED:
        if allow_no_lock:
            return
        raise DidNotHaveInstanceLockError()
    
    os.makedirs(os.path.dirname(LOCK_FILE_PATH), exist_ok=True)
    
    with FileLock(LOCK_FILE_PATH, timeout=timeout, mode=_FILE_LOCK_MODE):
        # Make sure we have the lock
        if not os.path.exists(PID_FILE_PATH) or _parse_pid_from_pid_file() != os.getpid():
            raise DidNotHaveInstanceLockError()
        
        # We have it, remove the PID file
        os.remove(PID_FILE_PATH)
        __LOCK_ACQUIRED = False
        info("Successfully released robot instance lock")


def is_robot_instance_locked() -> bool:
    """Check if the robot instance lock is held by the current process"""
    return __LOCK_ACQUIRED


def _is_process_running(pid: int) -> bool:
    """Check if a process is actually running (not a zombie)."""
    if not psutil.pid_exists(pid):
        return False
    
    try:
        # Check if process is a zombie - zombies are considered "dead" for our purposes
        if psutil.Process(pid).status() == psutil.STATUS_ZOMBIE:
            return False
        return True
    except psutil.NoSuchProcess:  # Process doesn't exist, shouldn't ever happen since we check pid_exists first
        return False
    except psutil.AccessDenied:  # Can't access process info, but it exists - assume it's running
        return True


def _parse_pid_from_pid_file() -> int:
    """Get the PID of the running robot instance. Assumes file exists and we have the file lock"""
    with open(PID_FILE_PATH, 'r') as f:
        file_str = f.read()
    try:
        return int(file_str.strip())
    except ValueError:
        raise InvalidPIDFileError(PID_FILE_PATH)


def check_still_own_lock() -> bool:
    """Check if the current process still owns the lock"""
    try:
        return __LOCK_ACQUIRED and _parse_pid_from_pid_file() == os.getpid()
    except Exception:
        return False



import time
from feathersdk.utils.common import timediff, currtime

class PerfLogger:
    def __init__(self):
        # Use perf_counter for performance measurements - higher precision and not affected by system clock adjustments
        self.last_time = currtime()
    
    def log(self, msg=""):
        # Use perf_counter for performance measurements - higher precision and not affected by system clock adjustments
        delta = timediff(self.last_time)
        print(f"{msg} Δt: {delta:.6f}s")
        self.last_time = currtime()


"""Data classes and enums for the Robstride motor parameters/types"""
from functools import cached_property
from feathersdk.utils.feathertypes import *
from dataclasses import dataclass, field
from feathersdk.comms.system import *
from enum import IntEnum, Enum
import math
from feathersdk.utils.common import make_exception_pickleable
from feathersdk.utils.logger import warning

if TYPE_CHECKING:
    from .robstride_motor import RobstrideMotor


class RobstrideVersion(Enum):
    RS00 = 0x00
    RS01 = 0x01
    RS02 = 0x02
    RS03 = 0x03
    RS04 = 0x04
    RS05 = 0x05
    RS06 = 0x06


RobstrideVersionType = Union[int, str, RobstrideVersion]


def get_robstride_version(ver: RobstrideVersionType) -> RobstrideVersion:
    """Get the RobstrideVersion enum from the input. Can pass int, str, or RobstrideVersion.
    
    EG: 2, '4', RobstrideVersion.RS02, 'RS04'
    """
    if isinstance(ver, RobstrideVersion):
        return ver
    if isinstance(ver, int):
        return RobstrideVersion(ver)
    if isinstance(ver, str):
        try:
            int_ver = int(ver)
        except ValueError:
            int_ver = None
        if int_ver is not None:
            return RobstrideVersion(int_ver)
        else:
            ver_str = ver.lower().replace("robstrideversion.", "")
            for v in RobstrideVersion:
                if v.name.lower() == ver_str:
                    return v
            raise ValueError(f"Invalid version string: \"{ver}\"")
    raise TypeError(f"Invalid version type: {type(ver)}")


class RobstrideRunMode(IntEnum):
    Operation = 0
    Position = 1
    Speed = 2
    Current = 3


class RobstrideMotorError(Enum):
    """Errors that can be returned in the CAN ID of a feedback message.
    
    Positive values are for motor feedback errors, negative values are for fault feedback errors. All errors 
    are powers of 2. So abs(value) is the error bit position.
    """
    UnderVoltage = 0
    OverCurrent = 1
    OverTemperature = 2
    MagneticEncoding = 3
    GridlockOverload = 4

    # Not in the CAN ID, but in the fault feedback message
    DriverChipFault = 5
    OverVoltage = 6
    EncoderNotCalibrated = 7

    @staticmethod
    def _motor_error_bits() -> Dict[int, 'RobstrideMotorError']:
        return {
            1 << 0: RobstrideMotorError.UnderVoltage, 
            1 << 1: RobstrideMotorError.OverCurrent, 
            1 << 2: RobstrideMotorError.OverTemperature, 
            1 << 3: RobstrideMotorError.MagneticEncoding, 
            1 << 4: RobstrideMotorError.GridlockOverload
        }

    @staticmethod
    def _fault_error_bits() -> Dict[int, 'RobstrideMotorError']:
        return {
            1 << 0: RobstrideMotorError.OverTemperature,
            1 << 1: RobstrideMotorError.DriverChipFault,
            1 << 2: RobstrideMotorError.UnderVoltage,
            1 << 3: RobstrideMotorError.OverVoltage,
            1 << 7: RobstrideMotorError.EncoderNotCalibrated,
            1 << 14: RobstrideMotorError.GridlockOverload,
        }

    @staticmethod
    def _parse_bits(value: int, bits: Dict[int, 'RobstrideMotorError'], name: str) -> Set['RobstrideMotorError']:
        ret = set(error for bit, error in bits.items() if bit & value)

        if value < 0 or value & ~sum(bits.keys()) != 0:
            warning(f"{name} value {value} has unknown error bits, or is negative")
        
        return ret

    @staticmethod
    def parse_motor_feedback_errors(extra_byte: int) -> Set['RobstrideMotorError']:
        """Parse the errors from the extra byte of the can_id of a feedback message."""
        return RobstrideMotorError._parse_bits(extra_byte, RobstrideMotorError._motor_error_bits(), "Motor feedback")
    
    @staticmethod
    def parse_fault_feedback_errors(fault_bytes: int) -> Set['RobstrideMotorError']:
        """Parse the fault feedback errors from the fault feedback message."""
        return RobstrideMotorError._parse_bits(fault_bytes, RobstrideMotorError._fault_error_bits(), "Fault feedback")
    
    @staticmethod
    def errors_to_int(errors: Iterable['RobstrideMotorError']) -> int:
        """Convert an iterable of errors to an integer with the bits set for the errors."""
        return sum(1 << error.value for error in errors)


class RobstrideMotorMode(Enum):
    """Motor modes that can be returned in the CAN ID of a feedback message."""
    Reset = 0
    Calibration = 1
    Run = 2
    Unknown = 3

    @staticmethod
    def parse_mode(extra_byte: int) -> 'RobstrideMotorMode':
        """Parse the mode from the extra byte of the can_id of a feedback message."""
        if (ret := RobstrideMotorMode((extra_byte & 0b1100_0000) >> 6)) is RobstrideMotorMode.Unknown:
            warning(f"Unknown Robstride motor mode parsed: {extra_byte}")
        return ret


class RobstrideMotorMsg(Enum):
    GetDeviceId = 0x00
    OperationControl = 0x01
    MotorFeedback = 0x02
    Enable = 0x03
    Disable = 0x04
    ZeroPos = 0x06
    SetID = 0x07
    BeginFlash = 0x0b
    FlashFileInfo = 0x0c
    FlashData = 0x0d
    EndFlash = 0x0e
    ReadParam = 0x11
    WriteParam = 0x12
    FaultFeedback = 0x15
    SaveMotorData = 0x16
    SetMotorBaudRate = 0x17
    SetMotorActiveReporting = 0x18
    SetMotorProtocol = 0x19


class RobstrideFeedbackMessageBounds:
    """Bounds for the feedback message."""
    ANGLE_MAP = {v: FloatRange(-4.0 * math.pi, 4.0 * math.pi) for v in RobstrideVersion}
    VELOCITY_MAP = {
        RobstrideVersion.RS00: FloatRange(-33.0, 33.0),
        RobstrideVersion.RS01: FloatRange(-44.0, 44.0),
        RobstrideVersion.RS02: FloatRange(-44.0, 44.0),
        RobstrideVersion.RS03: FloatRange(-20.0, 20.0),
        RobstrideVersion.RS04: FloatRange(-15.0, 15.0),
        RobstrideVersion.RS05: FloatRange(-50.0, 50.0),
        RobstrideVersion.RS06: FloatRange(-50.0, 50.0),
    }
    TORQUE_MAP = {
        RobstrideVersion.RS00: FloatRange(-14.0, 14.0),
        RobstrideVersion.RS01: FloatRange(-17.0, 17.0),
        RobstrideVersion.RS02: FloatRange(-17.0, 17.0),
        RobstrideVersion.RS03: FloatRange(-60.0, 60.0),
        RobstrideVersion.RS04: FloatRange(-120.0, 120.0),
        RobstrideVersion.RS05: FloatRange(-5.5, 5.5),
        RobstrideVersion.RS06: FloatRange(-36.0, 36.0),
    }
    TEMP_DIVISOR = 10  # Same for all versions


class RobstrideOperationCommand:
    """Command builder for Robstride motor operation control messages.
    
    Args:
        motor: The RobstrideMotor instance to send the command to.
        t_torque: Target torque in Nm. Must be within version-specific torque bounds.
        t_angle: Target angle in radians. Must be within [-4π, 4π].
        t_vel: Target velocity in rad/s. Must be within version-specific velocity bounds.
        kp: Proportional gain. Must be within version-specific KP bounds.
        kd: Derivative gain. Must be within version-specific KD bounds.
    """

    KP_BOUNDS_MAP = {
        RobstrideVersion.RS00: FloatRange(0.0, 500.0),
        RobstrideVersion.RS01: FloatRange(0.0, 500.0),
        RobstrideVersion.RS02: FloatRange(0.0, 500.0),
        RobstrideVersion.RS03: FloatRange(0.0, 5000.0),
        RobstrideVersion.RS04: FloatRange(0.0, 5000.0),
        RobstrideVersion.RS05: FloatRange(0.0, 500.0),
        RobstrideVersion.RS06: FloatRange(0.0, 5000.0),
    }

    KD_BOUNDS_MAP = {
        RobstrideVersion.RS00: FloatRange(0.0, 5.0),
        RobstrideVersion.RS01: FloatRange(0.0, 5.0),
        RobstrideVersion.RS02: FloatRange(0.0, 5.0),
        RobstrideVersion.RS03: FloatRange(0.0, 100.0),
        RobstrideVersion.RS04: FloatRange(0.0, 100.0),
        RobstrideVersion.RS05: FloatRange(0.0, 5.0),
        RobstrideVersion.RS06: FloatRange(0.0, 100.0),
    }

    def __init__(self, motor: 'RobstrideMotor', t_torque: float, t_angle: float, t_vel: float, kp: float, kd: float):
        self.motor: 'RobstrideMotor' = motor
        self.target_torque = t_torque
        self.target_angle = t_angle
        self.target_velocity = t_vel
        self.kp = kp
        self.kd = kd
    
    def _validate(self):
        for name, value, bounds in [
            ("target_torque", self.target_torque, RobstrideFeedbackMessageBounds.TORQUE_MAP[self.motor.version]), 
            ("target_angle", self.target_angle, RobstrideFeedbackMessageBounds.ANGLE_MAP[self.motor.version]), 
            ("target_velocity", self.target_velocity, RobstrideFeedbackMessageBounds.VELOCITY_MAP[self.motor.version]), 
            ("kp", self.kp, RobstrideOperationCommand.KP_BOUNDS_MAP[self.motor.version]), 
            ("kd", self.kd, RobstrideOperationCommand.KD_BOUNDS_MAP[self.motor.version])
        ]:
            if value < bounds[0] or value > bounds[1]:
                raise ValueError(f"{name} value ({value}) is out of bounds [{bounds[0]}, {bounds[1]}]")

    def build_command(self) -> Tuple[int, bytes]:
        """Build the command bytes for the operation control message."""
        self._validate()
        
        version = self.motor.version
        t_bytes = S_uint16be.map_and_pack(self.target_torque, *RobstrideFeedbackMessageBounds.TORQUE_MAP[version])
        a_bytes = S_uint16be.map_and_pack(self.target_angle, *RobstrideFeedbackMessageBounds.ANGLE_MAP[version])
        v_bytes = S_uint16be.map_and_pack(self.target_velocity, *RobstrideFeedbackMessageBounds.VELOCITY_MAP[version])
        kp_bytes = S_uint16be.map_and_pack(self.kp, *RobstrideOperationCommand.KP_BOUNDS_MAP[version])
        kd_bytes = S_uint16be.map_and_pack(self.kd, *RobstrideOperationCommand.KD_BOUNDS_MAP[version])
        
        data = a_bytes + v_bytes + kp_bytes + kd_bytes
        can_id = self.motor._msg_can_id(RobstrideMotorMsg.OperationControl, self.motor.motor_id)
        return ((can_id & 0xFF00_00FF) | (t_bytes[0] << 16) | (t_bytes[1] << 8), data)


# Type for the union of all possible parameter value types
RobstrideParamValue = Union[float, int, RobstrideRunMode]
T = TypeVar("T", bound=RobstrideParamValue)


@dataclass
class RobstrideParam(Generic[T]):
    id: int
    mode: Literal["r", "w", "rw"]
    dtype: StructType
    default: Dict[RobstrideVersion, T]
    range: Dict[RobstrideVersion, FloatRange]
    name: str = field(default="")
    value: T = field(default=None)

    def __post_init__(self):
        for version in RobstrideVersion:
            if version not in self.range:
                raise KeyError(f"Version {version} not found in range for parameter {self.name}")
            if self.range[version].min > self.range[version].max:
                raise ValueError(f"Invalid range {self.range[version]} for parameter {self.name}")
        assert self.default is not None, f"Default value is required"
        self.value = self.default[RobstrideVersion.RS00]
    
    def is_within_range(self, value: float, version: RobstrideVersionType) -> bool:
        """Checks if value is within the acceptible range for the parameter for the given robstride motor version"""
        version = get_robstride_version(version)
        return self.range[version].min <= value <= self.range[version].max
    
    @property
    def readable(self) -> bool:
        return self.mode in ["r", "rw"]
    
    @property
    def writable(self) -> bool:
        return self.mode in ["w", "rw"]
    
    def unpack_and_clamp(self, data: bytes, version: RobstrideVersionType) -> T:
        """Unpack the data and clamp it to the range for the parameter."""
        return self.default[version].__class__(self.range[version].clamp(self.dtype.unpack(data)))


_CURRENT_LIMIT_MAP = {
    RobstrideVersion.RS00: 16.0,
    RobstrideVersion.RS01: 23.0,
    RobstrideVersion.RS02: 16.0,
    RobstrideVersion.RS03: 43.0,
    RobstrideVersion.RS04: 90.0,
    RobstrideVersion.RS05: 11.0,
    RobstrideVersion.RS06: 57.0,
}

_SPEED_LIMIT_MAP = {
    RobstrideVersion.RS00: 33.0,
    RobstrideVersion.RS01: 44.0,
    RobstrideVersion.RS02: 33.0,
    RobstrideVersion.RS03: 20.0,
    RobstrideVersion.RS04: 20.0,
    RobstrideVersion.RS05: 50.0,
    RobstrideVersion.RS06: 50.0,
}


# Used to mark a (-inf, inf) range as one that is unknown, as opposed to a testing infinity range
_NO_KNOWN_RANGE = {v: FloatRange(float('-inf'), float('inf')) for v in RobstrideVersion}

def _vrange(
    _min: Optional[Union[float, Dict[RobstrideVersion, float]]] = None,
    _max: Optional[Union[float, Dict[RobstrideVersion, float]]] = None
) -> Dict[RobstrideVersion, FloatRange]:
    """Helper function to create a range for a parameter.
    Pass floats for that range for all versions, or a single dictionary of version -> value for the range. If that
    dictionary is passed in _min, will use range (-d[v], d[v]), otherwise if passed in _max, will use range (0, d[v]).
    """
    if isinstance(_min, (float, int)):
        return {v: FloatRange(float(_min), float(_max)) for v in RobstrideVersion}
    elif _min is not None:
        assert isinstance(_min, dict) and _max is None, f"Bad args: {_min}, {_max}"
        return {v: FloatRange(float(-_min[v]), float(_min[v])) for v in RobstrideVersion}
    elif _max is not None:
        assert isinstance(_max, dict) and _min is None, f"Bad args: {_min}, {_max}"
        return {v: FloatRange(0.0, float(_max[v])) for v in RobstrideVersion}
    else:
        raise TypeError(f"Bad args: {_min}, {_max}")


def _vdef(default: T) -> Dict[RobstrideVersion, T]:
    """Helper function to create a default value for a parameter for all versions"""
    return {v: default for v in RobstrideVersion}


@dataclass
class RobstrideParamList:
    """Contains all the parameters for all robstride versions"""
    
    run_mode: RobstrideParam[RobstrideRunMode] = field(default_factory=lambda: RobstrideParam(
        id=0x7005, mode="rw", dtype=S_uint8le, default=_vdef(RobstrideRunMode.Operation), range=_vrange(0, 3),
    ))
    "Run mode selector (0: operation, 1: position PP, 2: velocity, 3: current)"
    iq_ref: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x7006, mode="rw", dtype=S_float32le, default=_vdef(0.0), range=_vrange(_CURRENT_LIMIT_MAP),
    ))
    "Current mode Iq command (A)"
    spd_ref: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x700A, mode="rw", dtype=S_float32le, default=_vdef(0.0), range=_vrange(_SPEED_LIMIT_MAP),
    ))
    "Velocity mode rotational speed command (rad/s)"
    limit_torque: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x700B, mode="rw", dtype=S_float32le, default=_vdef(5.0),
        range={
            RobstrideVersion.RS00: FloatRange(0, RobstrideFeedbackMessageBounds.TORQUE_MAP[RobstrideVersion.RS00].max),
            RobstrideVersion.RS01: FloatRange(0, RobstrideFeedbackMessageBounds.TORQUE_MAP[RobstrideVersion.RS01].max),
            RobstrideVersion.RS02: FloatRange(0, 14.0),  # Don't know why this is apparantly different
            RobstrideVersion.RS03: FloatRange(0, RobstrideFeedbackMessageBounds.TORQUE_MAP[RobstrideVersion.RS03].max),
            RobstrideVersion.RS04: FloatRange(0, RobstrideFeedbackMessageBounds.TORQUE_MAP[RobstrideVersion.RS04].max),
            RobstrideVersion.RS05: FloatRange(0, RobstrideFeedbackMessageBounds.TORQUE_MAP[RobstrideVersion.RS05].max),
            RobstrideVersion.RS06: FloatRange(0, RobstrideFeedbackMessageBounds.TORQUE_MAP[RobstrideVersion.RS06].max),
        }
    ))
    "Torque limit (Nm)"
    cur_kp: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x7010, mode="rw", dtype=S_float32le, default=_vdef(0.17), range=_vrange(float('-inf'), float('inf')),
    ))
    "Current loop Kp"
    cur_ki: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x7011, mode="rw", dtype=S_float32le, default=_vdef(0.012), range=_vrange(float('-inf'), float('inf')),
    ))
    "Current loop Ki"
    cur_filt_gain: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x7014, mode="rw", dtype=S_float32le, default=_vdef(0.1), range=_vrange(0, 1.0),
    ))
    "Current loop filter gain"
    loc_ref: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x7016, mode="rw", dtype=S_float32le, default=_vdef(0.0), range=_vrange(float('-inf'), float('inf')),
    ))
    "Position mode angle command (rad)"
    limit_spd: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x7017, mode="rw", dtype=S_float32le, default=_vdef(0.0), range=_vrange(None, _SPEED_LIMIT_MAP),
    ))
    "CSP mode speed limit (rad/s)"
    limit_cur: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x7018, mode="rw", dtype=S_float32le, default=_vdef(10.0), range=_vrange(None, _CURRENT_LIMIT_MAP),
    ))
    "Velocity/position mode current limit (A)"
    mech_pos: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x7019, mode="r", dtype=S_float32le, default=_vdef(0.0), range=_NO_KNOWN_RANGE,
    ))
    "Mechanical angle of the loading coil (rad)"
    iqf: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x701A, mode="r", dtype=S_float32le, default=_vdef(0.0), range=_vrange(_CURRENT_LIMIT_MAP),
    ))
    "Iq filter value (A)"
    mech_vel: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x701B, mode="r", dtype=S_float32le, default=_vdef(0.0), range=_vrange(_SPEED_LIMIT_MAP),
    ))
    "Load speed (rad/s)"
    vbus: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x701C, mode="r", dtype=S_float32le, default=_vdef(0.0), range=_NO_KNOWN_RANGE,
    ))
    "Bus voltage (V)"
    loc_kp: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x701E, mode="rw", dtype=S_float32le, range=_vrange(float('-inf'), float('inf')), default={
            RobstrideVersion.RS00: 40.0,
            RobstrideVersion.RS01: 40.0,
            RobstrideVersion.RS02: 40.0,
            RobstrideVersion.RS03: 60.0,
            RobstrideVersion.RS04: 60.0,
            RobstrideVersion.RS05: 40.0,
            RobstrideVersion.RS06: 40.0,
        },
    ))
    "Position loop Kp"
    spd_kp: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x701F, mode="rw", dtype=S_float32le, default=_vdef(6.0), range=_vrange(float('-inf'), float('inf')),
    ))
    "Velocity loop Kp"
    spd_ki: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x7020, mode="rw", dtype=S_float32le, default=_vdef(0.02), range=_vrange(float('-inf'), float('inf')),
    ))
    "Velocity loop Ki"
    spd_filt_gain: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x7021, mode="rw", dtype=S_float32le, default=_vdef(0.1), range=_vrange(float('-inf'), float('inf')),
    ))
    "Speed filter gain"
    acc_rad: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x7022, mode="rw", dtype=S_float32le, range=_vrange(float('-inf'), float('inf')), default={
            RobstrideVersion.RS00: 20.0,
            RobstrideVersion.RS01: 20.0,
            RobstrideVersion.RS02: 20.0,
            RobstrideVersion.RS03: 20.0,
            RobstrideVersion.RS04: 15.0,  # Why is this different???
            RobstrideVersion.RS05: 20.0,
            RobstrideVersion.RS06: 20.0,
        },
    ))
    "Velocity mode acceleration (rad/s^2)"
    vel_max: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x7024, mode="rw", dtype=S_float32le, default=_vdef(10.0), range=_NO_KNOWN_RANGE,
    ))
    "PP mode speed (rad/s)"
    acc_set: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x7025, mode="rw", dtype=S_float32le, default=_vdef(10.0), range=_NO_KNOWN_RANGE,
    ))
    "PP mode acceleration (rad/s^2)"
    epscan_time: RobstrideParam[int] = field(default_factory=lambda: RobstrideParam(
        id=0x7026, mode="rw", dtype=S_uint16le, default=_vdef(1), range=_vrange(*S_uint16le.range),
    ))
    "Active report interval (1 -> 10ms, +1 adds 5ms)"
    can_timeout: RobstrideParam[int] = field(default_factory=lambda: RobstrideParam(
        id=0x7028, mode="rw", dtype=S_uint32le, default=_vdef(0), range=_vrange(*S_uint32le.range),
    ))
    "CAN timeout threshold (20000 = 1s)"
    zero_sta: RobstrideParam[int] = field(default_factory=lambda: RobstrideParam(
        id=0x7029, mode="rw", dtype=S_uint8le, default=_vdef(0), range=_vrange(*S_uint8le.range),
    ))
    "Zero flag (0: [0, 2π], 1: [-pi, pi])"
    damper: RobstrideParam[int] = field(default_factory=lambda: RobstrideParam(
        id=0x702A, mode="rw", dtype=S_uint8le, default=_vdef(0), range=_vrange(*S_uint8le.range),
    ))
    "Damping switch"
    add_offset: RobstrideParam[float] = field(default_factory=lambda: RobstrideParam(
        id=0x702B, mode="rw", dtype=S_float32le, default=_vdef(0.0), range=_vrange(float('-inf'), float('inf')),
    ))
    "Zero offset"

    @cached_property
    def as_dict(self) -> Dict[str, RobstrideParam]:
        """Returns dictionary of parameter name to parameter"""
        return {name: param for name, param in self.__dict__.items() if isinstance(param, RobstrideParam)}
    
    def __getitem__(self, key: str) -> RobstrideParam:
        return self.as_dict[key.lower()]
    
    def __setitem__(self, key: str, value: RobstrideParamValue) -> None:
        self.as_dict[key.lower()].value = value
    
    def __iter__(self) -> Iterable[RobstrideParam]:
        return iter(self.as_dict.values())
    
    def __post_init__(self) -> None:
        for name, param in self.as_dict.items():
            param.name = name


_ROBSTRIDE_PARAMS = RobstrideParamList()
_ROBSTRIDE_PARAMS_BY_ID = {param.id: param for param in _ROBSTRIDE_PARAMS.as_dict.values()}


def get_param_info(param_id: Union[int, str, RobstrideParam]) -> RobstrideParam:
    """Get the parameter info for the given parameter ID or name.
    
    Args:
        param_id: The parameter ID or name. Case insensitive.
    """
    try:
        if isinstance(param_id, RobstrideParam):
            return param_id
        if isinstance(param_id, str):
            return _ROBSTRIDE_PARAMS[param_id.lower()]
        return _ROBSTRIDE_PARAMS_BY_ID[param_id]
    except KeyError:
        raise UnknownRobstrideParamError(param_id=param_id)


@make_exception_pickleable
class UnknownRobstrideParamError(Exception):
    """Raised when an unknown Robstride parameter is encountered."""
    def __init__(self, *, param_id: Union[int, str]):
        super().__init__(f"Unknown Robstride parameter: {param_id}")
        self.param_id = param_id



from dataclasses import dataclass
import numpy as np
import enum

import struct
import time
from typing import Any, List, Optional

import can


def uint_to_float(x_int: int, x_min: float, x_max: float, bits: int) -> float:
    """Converts unsigned int to float, given range and number of bits."""
    span = x_max - x_min
    return (x_int * span / ((1 << bits) - 1)) + x_min


def float_to_uint(x: float, x_min: float, x_max: float, bits: int) -> int:
    """Converts a float to an unsigned int, given range and number of bits."""
    span = x_max - x_min
    x = min(x, x_max)
    x = max(x, x_min)
    return int((x - x_min) * ((1 << bits) - 1) / span)

@dataclass
class MotorConstants:
    POSITION_MAX: float = 12.5
    POSITION_MIN: float = -12.5

    VELOCITY_MAX: float = 45
    VELOCITY_MIN: float = -45

    TORQUE_MAX: float = 54
    TORQUE_MIN: float = -54

    ####### Mihgt be used for other motors #######
    CURRENT_MAX: float = 1.0
    CURRENT_MIN: float = -1.0
    KT:float = 1.0
    ##############################

    KP_MAX: float = 500.0
    KP_MIN: float = 0.0
    KD_MAX: float = 5.0
    KD_MIN: float = 0.0

@dataclass
class MotorInfo:
    """Class to represent motor information.

    Attributes:
        id (int): Motor ID.
        target_torque (int): Target torque value.
        vel (float): Motor speed.
        eff (float): Motor current.
        pos (float): Encoder value.
        voltage (float): Motor voltage.
        temperature (float): Motor temperature.

    """

    id: int
    error_code: int
    target_torque: int = 0
    vel: float = 0.0
    eff: float = 0
    pos: float = 0
    voltage: float = -1
    temp_mos: float = -1
    temp_rotor: float = -1


@dataclass
class FeedbackFrameInfo:
    id: int
    error_code: int
    error_message: str
    position: float
    velocity: float
    torque: float
    temperature_mos: float
    temperature_rotor: float


@dataclass
class EncoderInfo:
    encoder = -1
    encoder_raw = -1
    encoder_offset = -1


class MotorErrorCode:
    disabled = 0x0
    normal = 0x1
    over_voltage = 0x8
    under_voltage = 0x9
    over_current = 0xA
    mosfet_over_temperature = 0xB
    motor_over_temperature = 0xC
    loss_communication = 0xD
    overload = 0xE

    # create a dict map error code to error message
    motor_error_code_dict = {
        normal: "normal",
        disabled: "disabled",
        over_voltage: "over voltage",
        under_voltage: "under voltage",
        over_current: "over current",
        mosfet_over_temperature: "mosfet over temperature",
        motor_over_temperature: "motor over temperature",
        loss_communication: "loss communication",
        overload: "overload",
    }
    # covert to decimal
    motor_error_code_dict = {int(k): v for k, v in motor_error_code_dict.items()}

    @classmethod
    def get_error_message(cls, error_code: int) -> str:
        return cls.motor_error_code_dict.get(int(error_code), f"Unknown error code: {error_code}")

class MotorType:
    DM4310 = "DM4310"

    @classmethod
    def get_motor_constants(cls, motor_type: str = None) -> MotorConstants:
        """Get motor constants for DM4310.
        
        Args:
            motor_type: Optional motor type (ignored, kept for compatibility).
        
        Returns:
            MotorConstants for DM4310.
        """
        return MotorConstants(
            POSITION_MAX=12.5,
            POSITION_MIN=-12.5,
            VELOCITY_MAX=30,
            VELOCITY_MIN=-30,
            TORQUE_MAX=7,
            TORQUE_MIN=-7,
            KP_MAX=500.0,
            KP_MIN=0.0,
            KD_MAX=5.0,
            KD_MIN=0.0,
        )


class AutoNameEnum(enum.Enum):
    def _generate_next_value_(name: str, start: int, count: int, last_values: List[str]) -> str:
        return name

class ReceiveMode(AutoNameEnum):
    p16 = enum.auto()
    same = enum.auto()
    zero = enum.auto()
    plus_one = enum.auto()

    def get_receive_id(self, motor_id: int) -> int:
        if self == ReceiveMode.p16:
            return motor_id + 16
        elif self == ReceiveMode.same:
            return motor_id
        elif self == ReceiveMode.zero:
            return 0
        elif self == ReceiveMode.plus_one:
            return motor_id + 1
        else:
            raise NotImplementedError(f"receive_mode: {self} not recognized")

    def to_motor_id(self, receive_id: int) -> int:
        if self == ReceiveMode.p16:
            return receive_id - 16
        elif self == ReceiveMode.same:
            return receive_id
        elif self == ReceiveMode.zero:
            return 0
        else:
            raise NotImplementedError(f"receive_mode: {self} not recognized")

class RawCanInterface:
    def __init__(
        self,
        channel: str = "PCAN_USBBUS1",
        bustype: str = "pcan",
        bitrate: int = 1000000,
        name: str = "default_can_interface",
    ):
        self.bus = can.interface.Bus(bustype=bustype, channel=channel, bitrate=bitrate)
        self.busstate = self.bus.state
        self.name = name

    def close(self) -> None:
        """Shut down the CAN bus."""
        self.bus.shutdown()

    def _send_message_get_response(self, id: int, data: List[int], max_retry: int = 20) -> can.Message:
        self.try_receive_message(id)
        message = can.Message(arbitration_id=id, data=data, is_extended_id=False)
        for _ in range(max_retry):
            try:
                self.bus.send(message)
                response = self._receive_message(data[0])
                return response
            except (can.CanError, AssertionError) as e:
                print(e)
                # print warning in red
                print(
                    "\033[91m"
                    + f"CAN Error {self.name}: Failed to communicate with motor {data[0]} over can bus. Retrying..."
                    + "\033[0m"
                )
            time.sleep(0.002)
        raise AssertionError(
            f"fail to communicate with the motor {id} on {self.name} at can channel {self.bus.channel_info}"
        )

    def try_receive_message(self, motor_id: Optional[int] = None, timeout: float = 0.009) -> can.Message:
        """Try to receive a message from the CAN bus.

        Args:
            timeout (float): The time to wait for a message (in seconds).

        Returns:
            can.Message: The received message, or None if no message is received.
        """
        try:
            return self._receive_message(motor_id, timeout)
        except AssertionError:
            return None

    def _receive_message(self, motor_id: Optional[int] = None, timeout: float = 0.009) -> can.Message:
        """Receive a message from the CAN bus.

        Args:
            timeout (float): The time to wait for a message (in seconds).

        Returns:
            can.Message: The received message.

        Raises:
            AssertionError: If no message is received within the timeout.
        """
        start_time = time.time()
        while (time.time() - start_time) < timeout:
            message = self.bus.recv(timeout=0.002)
            if message:
                return message
        raise AssertionError(
            f"Failed to receive message, {self.name} motor id {motor_id} motor timeout. Check if the motor is powered on or if the motor ID exists."
        )


def bytes_to_uint32(data: bytearray) -> int:
    """Convert the last 4 bytes (data[4:8]) in a bytearray to uint32 (little-endian)."""
    return struct.unpack("<I", data[4:8])[0]


def bytes_to_float32(data: bytearray) -> float:
    """Convert the last 4 bytes (data[4:8]) in a bytearray to float32 (little-endian)."""
    return struct.unpack("<f", data[4:8])[0]


def uint32_to_bytes(value: int) -> bytearray:
    """Convert a uint32 value to a 4-byte bytearray (little-endian).

    Args:
        value (int): A uint32 value to be converted to bytes.

    Returns:
        bytearray: The corresponding 4-byte representation in little-endian order.
    """
    return bytearray(struct.pack("<I", value))


def float32_to_bytes(value: float) -> bytearray:
    """Convert a float32 value to a 4-byte bytearray (little-endian).

    Args:
        value (float): A float32 value to be converted to bytes.

    Returns:
        bytearray: The corresponding 4-byte representation in little-endian order.
    """
    return bytearray(struct.pack("<f", value))


register_addr_map = {
    "KT_value": (1, bytes_to_float32),
    "OT_value": (2, bytes_to_float32),
    "master_id": (7, bytes_to_uint32),
    "id": (8, bytes_to_uint32),
    "timeout": (9, bytes_to_uint32),
    "inertia": (12, bytes_to_float32),
    "sw_ver": (14, bytes_to_uint32),
    "flux": (19, bytes_to_float32),
    "gear_ratio": (20, bytes_to_float32),
    "gear_eff": (30, bytes_to_float32),
}

register_info_map = {"control_mode": {1: "MIT", 2: "pos_speed", 3: "speed", 4: "torque_pos"}}

def get_special_message_response(can_interface: RawCanInterface, motor_id: int, reg_name: str) -> Any:
    """Get the current value of a register from a motor.

    Args:
        can_interface (RawCanInterface): The CAN interface to use.
        motor_id (int): The ID of the motor to read the register value from.
        reg_name (str): The name of the register to read.

    Returns:
        Any: The value read from the register.
    """
    assert reg_name in register_addr_map, f"reg_name {reg_name} not in register_addr_map"
    reg_id, convert_func = register_addr_map[reg_name]
    for _ in range(3):
        try:
            message = can_interface._send_message_get_response(
                0x7FF, [motor_id, 0x00, 0x33, reg_id, 0x00, 0x00, 0x00, 0x00], max_retry=20
            )
            return convert_func(message.data)
        except Exception as e:
            can_interface.try_receive_message(motor_id)
    raise Exception(f"Failed to read {reg_name} of motor {motor_id} after 3 retries")


def write_special_message(can_interface: RawCanInterface, motor_id: int, reg_name: str, data: Any) -> Any:
    """Write a value to a register of a motor.

    Args:
        can_interface (RawCanInterface): The CAN interface to use.
        motor_id (int): The ID of the motor to write the register value for.
        reg_name (str): The name of the register to write.
        data (Any): The value to write to the register.

    Returns:
        Any: The value read from the register after writing.
    """
    assert reg_name in register_addr_map, f"reg_name {reg_name} not in register_addr_map"
    reg_id, convert_func = register_addr_map[reg_name]
    if convert_func is bytes_to_uint32:
        byte_list = uint32_to_bytes(data)
    elif convert_func is bytes_to_float32:
        byte_list = float32_to_bytes(data)
    for _ in range(3):
        try:
            message = can_interface._send_message_get_response(
                0x7FF, [motor_id, 0x00, 0x55, reg_id] + list(byte_list), max_retry=20
            )
            result = convert_func(message.data)
            return result
        except Exception as e:
            can_interface.try_receive_message(motor_id)


def save_to_memory(can_interface: RawCanInterface, motor_id: int, reg_name: str) -> can.Message:
    """Save the current value of a register to the motor's memory.

    Args:
        can_interface (RawCanInterface): The CAN interface to use.
        motor_id (int): The ID of the motor to save the register value for.
        reg_name (str): The name of the register to save.

    Returns:
        can.Message: The response message from the motor.
    """
    assert reg_name in register_addr_map, f"reg_name {reg_name} not in register_addr_map"
    reg_id, convert_func = register_addr_map[reg_name]
    message = can_interface._send_message_get_response(
        0x7FF, [motor_id, 0x00, 0xAA, reg_id, 0x00, 0x00, 0x00, 0x00], max_retry=20
    )
    return message




import math
from .steppable_system import SteppableSystem
from .motors.motors_manager import MotorsManager, SafeStopPolicy
from .motors.robstride import RobstrideMotor

class TestSystem(SteppableSystem):
    def __init__(self, motors_manager: MotorsManager):
        motors_map = {
            "test_motor_1": RobstrideMotor("01", "test_motor_1", {
                # "joint_limits": [-3.14, 3.14],
                "max_torque": 0.75,
                "max_velocity": 3.14,
                # "max_position_dx": 0.9
            })
        }
        self.motors_manager = motors_manager
        self.motors_manager.add_motors(motors_map, "test", {"safe_stop_policy": SafeStopPolicy.COMPLIANCE_MODE})
        self.motors_manager.find_motors()
        self.motors = motors_map

    def _step(self):
        pass



"""
Pinocchio model loader for robots.

This module provides PinocchioModel class for loading and managing Pinocchio models
and collision geometry, making them available for multiple tasks (kinematics, dynamics,
collision checking, visualization, etc.).
"""

import logging
import pinocchio as pin
import numpy as np
from pathlib import Path
from feathersdk.utils.feathertypes import List, Optional

logger = logging.getLogger(__name__)


class PinocchioModel:
    """
    Reusable Pinocchio model and geometry loader.
    
    This class loads and manages the Pinocchio model and collision geometry,
    making them available for multiple tasks (kinematics, dynamics, collision checking, etc.).
    """
    
    def __init__(
        self, 
        urdf_path: str, 
        package_dir: Optional[str] = None,
        load_collision: bool = True
    ):
        """
        Initialize Pinocchio model from URDF.
        
        Args:
            urdf_path: Path to the URDF file
            package_dir: Directory containing mesh files (for package:// paths)
            load_collision: If True, load collision geometry model
        """
        self.urdf_path = Path(urdf_path)
        
        # Determine package directory
        if package_dir is None:
            # Try to auto-detect: look for package.xml to find the package root
            current = self.urdf_path.parent
            while current != current.parent:  # Stop at filesystem root
                if (current / "package.xml").exists():
                    # Found package root (Nov11_description/)
                    # package_dir should be parent of this
                    self.package_dir = current.parent
                    logger.info(f"Auto-detected package_dir: {self.package_dir}")
                    break
                current = current.parent
            else:
                # Fallback: assume package is 2 levels up from URDF
                self.package_dir = self.urdf_path.parent.parent.parent
                logger.info(f"Using inferred package_dir: {self.package_dir}")
        else:
            self.package_dir = Path(package_dir)
        
        # Load the kinematic model
        self.model = pin.buildModelFromUrdf(str(self.urdf_path))
        self.data = self.model.createData()
        
        # Load collision geometry if requested
        self.geom_model = None
        self.geom_data = None
        if load_collision:
            self._load_collision_geometry()
        
        logger.info(f"✓ Model loaded: {self.model.name}")
        n_frames = getattr(self.model, 'nframes', len(self.model.names))
        logger.info(f"✓ Number of frames: {n_frames}")
        logger.info(f"✓ Number of joints: {self.model.njoints}")
        if self.geom_model:
            logger.info(f"✓ Number of collision geometries: {len(self.geom_model.geometryObjects)}")
    
    def _load_collision_geometry(self):
        """Load collision geometry from URDF."""
        logger.info(f"Loading collision geometry with package_dirs: [{self.package_dir}]")
        
        try:
            self.geom_model = pin.buildGeomFromUrdf(
                self.model,
                str(self.urdf_path),
                pin.GeometryType.COLLISION,
                None,  # geometry_model=None
                [str(self.package_dir)]  # package_dirs as list
            )
            self.geom_data = self.geom_model.createData()
            logger.info(f"✓ Successfully loaded collision geometry")
        except Exception as e:
            # Fallback: try without package_dirs
            logger.warning(f"⚠ Loading with package_dirs failed, trying without...")
            try:
                self.geom_model = pin.buildGeomFromUrdf(
                    self.model,
                    str(self.urdf_path),
                    pin.GeometryType.COLLISION
                )
                self.geom_data = self.geom_model.createData()
                logger.info(f"✓ Successfully loaded collision geometry (without package_dirs)")
            except Exception as e2:
                logger.error(f"\n❌ Failed to load collision geometry")
                logger.error(f"   package_dirs tried: [{self.package_dir}]")
                logger.error(f"   Errors: {e}, {e2}")
                raise RuntimeError(
                    f"Failed to load collision geometry. "
                    f"package_dirs: [{self.package_dir}]. "
                    f"Last error: {e2}"
                )
        
        # Initialize geometry data with neutral configuration
        try:
            q_neutral = pin.neutral(self.model)
            pin.forwardKinematics(self.model, self.data, q_neutral)
            pin.updateGeometryPlacements(
                self.model, self.data, self.geom_model, self.geom_data, q_neutral
            )
            logger.info("✓ Geometry data initialized successfully")
        except Exception as e:
            logger.warning(f"⚠ Warning: Could not initialize geometry data: {e}")
    
    def update_kinematics(self, q: np.ndarray):
        """
        Update forward kinematics for the given configuration.
        
        Args:
            q: Joint configuration vector (nq x 1)
        """
        if q.ndim > 1:
            q = q.flatten()
        q = np.asarray(q, dtype=np.float64)
        
        if len(q) != self.model.nq:
            raise ValueError(
                f"Configuration size mismatch: got {len(q)}, expected {self.model.nq}"
            )
        
        pin.forwardKinematics(self.model, self.data, q)
    
    def update_geometry(self, q: np.ndarray):
        """
        Update geometry placements based on joint configuration.
        
        Args:
            q: Joint configuration vector (nq x 1)
        """
        if self.geom_model is None:
            raise RuntimeError("Collision geometry not loaded. Set load_collision=True in __init__")
        
        self.update_kinematics(q)
        pin.updateGeometryPlacements(
            self.model, self.data, self.geom_model, self.geom_data, q
        )
    
    def get_link_names(self) -> List[str]:
        """Get list of all link names in the model."""
        n_frames = getattr(self.model, 'nframes', len(self.model.names))
        return [self.model.names[i] for i in range(n_frames)]
    
    def get_geometry_names(self) -> List[str]:
        """Get list of all collision geometry names."""
        if self.geom_model is None:
            return []
        return [geom.name for geom in self.geom_model.geometryObjects]


def load_pinocchio_model_for_nov11(
    base_path: Optional[str] = None,
    urdf_filename: str = "Nov11_convex.urdf"
) -> PinocchioModel:
    """
    Convenience function to load PinocchioModel for the Nov11 robot.
    
    Args:
        base_path: Base path to featheros directory. If None, tries to infer.
        urdf_filename: Name of URDF file (default: "Nov11_convex.urdf")
    
    Returns:
        PinocchioModel instance
    """
    if base_path is None:
        # Try to find featheros directory
        current = Path.cwd()
        if 'featheros' in str(current):
            base_path = str(current)[:str(current).find('featheros') + len('featheros')]
        else:
            base_path = str(current)
    
    urdf_path = Path(base_path) / "assets" / "feather_one" / "2025_11_11" / \
                "Nov11_description" / "urdf" / urdf_filename
    
    if not urdf_path.exists():
        raise FileNotFoundError(f"URDF not found at: {urdf_path}")
    
    # Package directory for meshes
    package_dir = urdf_path.parent.parent.parent  # Points to 2025_11_11/
    
    return PinocchioModel(str(urdf_path), package_dir=str(package_dir), load_collision=True)




from feathersdk.comms import CommsManager
import time
from typing import Optional


class InspireFingerJoint:
    """
    Represents a single finger joint in the RH56DFTP Inspire hand.
    """

    def __init__(self, motor_id: int, ip_address: str, motor_name: str, force_set_offset: int, default_max_force: int):
        """
        Initialize a finger joint.
        
        Args:
            motor_id: The motor ID for this finger joint
            ip_address: IP address of the hand controller (can include port, e.g., "192.168.11.210:6000")
            motor_name: Name of the motor
            force_set_offset: Offset for force set command (default: 12)
        """
        self.motor_id = motor_id
        self.motor_name = motor_name
        self.ip_address = ip_address
        self.force_set_offset = force_set_offset
        self.properties = {}
        self.comms = CommsManager()
        if self.comms.is_running():
            self.set_max_force(default_max_force)

    def move_percentage(self, percentage: float):
        """
        Move the finger joint. 0% is fully closed, 100% is fully open.
        
        Args:
            percentage: Percentage of movement (0.0 to 1.0)
        """
        self.comms.tcpsend_modbus(self.ip_address, 1, 255, 6, self.motor_id, int(percentage * 1000))
        time.sleep(0.01)

    def set_max_force(self, force: int):
        """
        Set the max force for the finger joint.
        
        Args:
            force: Maximum force value
        """
        # Convert motor_id from bytes to int, add offset, then convert back to bytes
        force_set_id = int(self.motor_id) + self.force_set_offset
        self.comms.tcpsend_modbus(self.ip_address, 1, 255, 6, force_set_id, force)
        time.sleep(0.005)


def create_finger_joints_from_config(config: dict, hand_side: str) -> dict:
    """
    Create finger joints from a config dictionary.
    
    Args:
        config: Configuration dictionary containing hand motor definitions
        hand_side: "right" or "left" to specify which hand
        force_set_offset: Offset for force set command (default: 12)
        
    Returns:
        Dictionary mapping motor names to InspireFingerJoint objects
    """
    force_set_offset = config.get("constants", {}).get("RH56DFTP_FORCE_OFFSET", 12)
    default_max_force = config.get("constants", {}).get("RH56DFTP_DEFAULT_MAX_FORCE", 500)
    hand_key = f"rh56dftp_hand_{hand_side}"
    if hand_key not in config:
        return {}
    
    hand_config = config[hand_key]
    if "motors" not in hand_config:
        return {}
    
    finger_joints = {}
    for motor_name, motor_info in hand_config["motors"].items():
        finger_joints[motor_name] = InspireFingerJoint(
            motor_id=motor_info["motor_id"],
            ip_address=motor_info["ip_address"],
            motor_name=motor_info["motor_name"],
            force_set_offset=force_set_offset,
            default_max_force=default_max_force
        )
        
    return finger_joints
